package edu.usc.ict.iago.agent;

import java.util.LinkedList;

import javax.websocket.Session;

import edu.usc.ict.iago.utils.BehaviorPolicy;
import edu.usc.ict.iago.utils.Event;
import edu.usc.ict.iago.utils.ExpressionPolicy;
import edu.usc.ict.iago.utils.GameSpec;
import edu.usc.ict.iago.utils.GeneralVH;
import edu.usc.ict.iago.utils.History;
import edu.usc.ict.iago.utils.MessagePolicy;
import edu.usc.ict.iago.utils.Offer;
import edu.usc.ict.iago.utils.Preference;
import edu.usc.ict.iago.utils.ServletUtils;

public class IAGOPinocchioVH extends GeneralVH {
	
	protected BehaviorPolicy behavior;
	protected ExpressionPolicy expression;
	protected MessagePolicy messages;
	protected GameSpec game;
	protected Offer lastOfferReceived;
	protected Offer lastOfferSent;
	protected AgentUtilsExtension utils;
	private boolean timeFlag = false;
		
	public IAGOPinocchioVH(GameSpec game, Session session, History history)
	{
		super("Pinocchio", session, history);
		this.game = game;
		utils = new AgentUtilsExtension(game, this);
		expression = new IAGONiceExpression(utils);
		messages = new IAGOPinocchioMessage(utils);
		behavior = new IAGONiceBehavior(utils);
		setHistory(new History());
		
	}
	
	@Override
	public LinkedList<Event> updateHistory(Event e)
	{
		//history.updateHistory(e);
		
		LinkedList<Event> resp = new LinkedList<Event>();
		if(e.getType().equals(Event.EventClass.SEND_EXPRESSION))
		{
			String expr = expression.getExpression(getHistory());
			Event e1 = new Event(History.VH_ID, Event.EventClass.SEND_EXPRESSION, expr, 2000, 0);
			//history.updateHistory(e1);
			resp.add(e1);
			
			Event e0 = null;
			
			if(e.getMessage().equals("sad") || e.getMessage().equals("angry"))
				e0 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, "I'm sorry, have I upset you?", 0);
			else if(e.getMessage().equals("happy"))
				e0 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, "I'm glad you're happy with how things are going!", 0);
			else if(e.getMessage().equals("surprised"))
				e0 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, "Have I said something unexpected?", 0);
			//history.updateHistory(e0);
			resp.add(e0);
			return resp;
		}
		
		//accepts anything better than batna
		if(e.getType().equals(Event.EventClass.FORMAL_ACCEPT))
		{
			Event lastOffer = utils.lastEvent(getHistory().getHistory(), Event.EventClass.SEND_OFFER);
			Event lastTime = utils.lastEvent(getHistory().getHistory(), Event.EventClass.TIME);
			
			int totalIssues = 0;
			for (int i = 0; i < game.getNumIssues(); i++)
				totalIssues += game.getIssueQuants()[i];
			if(lastOffer != null && lastTime != null)
			{
				//approximation
				int fairSplit = ((game.getNumIssues() + 1) * totalIssues / 4);
				//down to the wire, accept anything better than batna
				if(utils.myActualOfferValue(lastOffer.getOffer()) > game.getVHBATNA() && Integer.parseInt(lastTime.getMessage()) + 30 > game.getTotalTime()) 
				{
					Event e0 = new Event(History.VH_ID, Event.EventClass.FORMAL_ACCEPT, 0);
					//history.updateHistory(e0);
					resp.add(e0);
					return resp;
				}
				//accept anything better than fair minus one issue
				if(utils.myActualOfferValue(lastOffer.getOffer()) > fairSplit - game.getNumIssues())
				{
					Event e0 = new Event(History.VH_ID, Event.EventClass.FORMAL_ACCEPT, 0);
					//history.updateHistory(e0);
					resp.add(e0);
					return resp;
				}
				else
				{
					Event e1 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, "I don't think I can agree with that offer.", 0);
					//history.updateHistory(e1);
					resp.add(e1);
					return resp;					
				}
			}
		}
		
		if(e.getType().equals(Event.EventClass.TIME))
		{
			boolean noResponse = true;
			int i = getHistory().getHistory().size() - 1 ;
			for(; i > 0 && i > getHistory().getHistory().size() - 6; i--)//if nothing from anyone for four time intervals
			{
				if(getHistory().getHistory().get(i).getType() != Event.EventClass.TIME)
					noResponse = false;
			}
			
			if(noResponse)
			{
				Event e1 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, "Can I provide more information to help us reach consensus?", 0);
				//history.updateHistory(e1);
				resp.add(e1);
			}
			
			//times up
			if(!timeFlag && game.getTotalTime() - Integer.parseInt(e.getMessage()) < 30)
			{
				timeFlag = true;
				Event e1 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, "We're almost out of time!  Accept this quickly!", 0);
				//history.updateHistory(e1);
				resp.add(e1);
				Event e2 = new Event(History.VH_ID, Event.EventClass.SEND_OFFER, ((IAGONiceBehavior)behavior).getFinalOffer(getHistory()));
				//history.updateHistory(e2);
				resp.add(e2);
				
			}
			return resp;
		}
		
		//slightly more complex; also sends a message with its offer, and rejects your previous
		if(e.getType().equals(Event.EventClass.SEND_OFFER))
		{
			ServletUtils.debugLog("Agent Normalized ordering: " + utils.getVHOrdering());
			ServletUtils.debugLog("Optimal ordering: " + utils.getMinimaxOrdering());
			
//			String expr = expression.getExpression(history);
//			Event eExpr = new Event(History.VH_ID, Event.EventClass.SEND_EXPRESSION, expr);
//			//history.updateHistory(eExpr);
//			resp.add(eExpr);
			
			Offer o = e.getOffer();
			this.lastOfferReceived = o;
			
			boolean localFair = false;
			boolean totalFair = false;
			
			Offer allocated = ((IAGONiceBehavior)behavior).getAllocated();
			//int oppCur = utils.opponentValue(o, utils.getMinimaxOrdering());
			//int oppPrev = utils.opponentValue(allocated, utils.getMinimaxOrdering());
			
			if(utils.myActualOfferValue(o) > utils.myActualOfferValue(allocated))//net positive
				if(utils.myActualOfferValue(o) - utils.myActualOfferValue(allocated) + game.getNumIssues() > utils.opponentValue(o, utils.getMinimaxOrdering()) - utils.opponentValue(allocated, utils.getMinimaxOrdering()))
					localFair = true;//offer improvment is within one max value item of the same for me and my opponent
			
			if(utils.myActualOfferValue(o) + game.getNumIssues() > utils.opponentValue(o, utils.getMinimaxOrdering()))
				totalFair = true;//total offer still fair
			
			if (localFair && !totalFair)
			{
				Event eExpr = new Event(History.VH_ID, Event.EventClass.SEND_EXPRESSION, "sad", 2000, 0);
				//history.updateHistory(eExpr);
				resp.add(eExpr);
				Event e0 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, "Unfortunately, I cannot accept.  But that's getting close to being fair.", 0);
				//history.updateHistory(e0);
				resp.add(e0);
				Event e1 = new Event(History.VH_ID, Event.EventClass.OFFER_IN_PROGRESS, 0);
				//history.updateHistory(e1);
				resp.add(e1);
				Event e2 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, messages.getProposalLang(getHistory(), game), 3000);
				//history.updateHistory(e2);
				resp.add(e2);
				Event e3 = new Event(History.VH_ID, Event.EventClass.SEND_OFFER, behavior.getNextOffer(getHistory()), 700);
				//history.updateHistory(e3);
				this.lastOfferSent = e3.getOffer();
				resp.add(e3);
			}
			else if(localFair && totalFair)
			{
				Event eExpr = new Event(History.VH_ID, Event.EventClass.SEND_EXPRESSION, "happy", 2000, 0);
				//history.updateHistory(eExpr);
				resp.add(eExpr);
				Event e0 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, messages.getVHAcceptLang(getHistory(), game), 0);
				//history.updateHistory(e0);
				resp.add(e0);
				((IAGONiceBehavior)behavior).updateAllocated(this.lastOfferReceived);
				if(utils.isFullOffer(o))
				{
					Event eFormal = new Event(History.VH_ID, Event.EventClass.FORMAL_ACCEPT, 0);
					//history.updateHistory(eFormal);
					resp.add(eFormal);
				}
			}
			else
			{
				Event eExpr = new Event(History.VH_ID, Event.EventClass.SEND_EXPRESSION, "sad", 2000, 0);
				//history.updateHistory(eExpr);
				resp.add(eExpr);
				Event e0 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, messages.getVHRejectLang(getHistory(), game), 0);
				//history.updateHistory(e0);
				resp.add(e0);
				Event e1 = new Event(History.VH_ID, Event.EventClass.OFFER_IN_PROGRESS, 0);
				//history.updateHistory(e1);
				resp.add(e1);
				Event e2 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, messages.getProposalLang(getHistory(), game), 3000);
				//history.updateHistory(e2);
				resp.add(e2);
				Event e3 = new Event(History.VH_ID, Event.EventClass.SEND_OFFER, behavior.getNextOffer(getHistory()), 700);
				//history.updateHistory(e3);
				this.lastOfferSent = e3.getOffer();
				resp.add(e3);
			}
			
			return resp;
		}
		
		if(e.getType().equals(Event.EventClass.SEND_MESSAGE))
		{
			Preference p = e.getPreference() == null ? null : new Preference(e.getPreference().getIssue1(), e.getPreference().getIssue2(), e.getPreference().getRelation(), e.getPreference().isQuery());
			if (p != null && !p.isQuery()) //a preference was expressed
			{
				utils.addPref(p);
				if(utils.reconcileContradictions())
				{
					//we simply drop the oldest expressed preferences until we are reconciled.  This is not the best method, as it may not be the the most efficient route.
					LinkedList<String> dropped = new LinkedList<String>();
					dropped.add(IAGOPinocchioMessage.prefToEnglish(utils.dequeuePref(), game));
					while(utils.reconcileContradictions())
						dropped.add(IAGOPinocchioMessage.prefToEnglish(utils.dequeuePref(), game));
					
					String drop = "";
					for (String s: dropped)
						drop += "\"" + s + "\", and ";
					
					drop = drop.substring(0, drop.length() - 7);
					
					Event e1 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, 
							"I'm sorry.  I must be misunderstanding.  Earlier, you said: \"" + drop + "\" Was that not correct?", 2000);
					//history.updateHistory(e1);
					resp.add(e1);
				}
			}
			
			
			Event e1 = new Event(History.VH_ID, Event.EventClass.SEND_MESSAGE, messages.getMessageResponse(getHistory(), game), 3000);
			//history.updateHistory(e1);
			resp.add(e1);
			
			
			if(e.getMessageCode() == 11)
			{
				Event e2 = new Event(History.VH_ID, Event.EventClass.SEND_OFFER, behavior.getNextOffer(getHistory()), 3000);
				//history.updateHistory(e2);
				this.lastOfferSent = e2.getOffer();
				resp.add(e2);
			}
			if(e.getMessageCode() == 101)//offer accepted
			{
				if(this.lastOfferSent != null)
					((IAGONiceBehavior)behavior).updateAllocated(this.lastOfferSent);
			}
			return resp;
		}
		
		return null;
	}

	@Override
	public String getArtName() {
		return "Brad";
	}
	
	@Override
	public String agentDescription() {
		return "<h1>Brad</h1><p>He is excited to begin negotiating!</p>";
	}
	

}